'''**Usage running functionalities.**
'''

import os
from functools import reduce
import numpy as np
import concurrent.futures as cf
from chem_gm.components.gmmEngine import createGMModel
from nntoolbox.filetoolbox import getDictFromGmx
from monal import monalconst as C
from monal.monalrecords import ModelData
from monal.driver import Driver, optim_leastsq
from monal.util.toolbox import maxWorkers
from nn1.NN1_API import nn1Error

# to-do : fonction get_result_from_smiles  averc parallelism
def get_results_from_smiles(modellist, inputs=None, level=0, 
                            modelindex=-1, parallel=True, progress=None):
    """Create graphmachine models and obtain their results.
    
    parameters:
     - modellist -> list of triplets (sourcefile, model name, smiles code) to perocess.
     - inputs -> None, vector or dictionary
         - None : no input values (default).
         - vector : vector of the input values, in the order of the inputs.
         - dictionary: input value for each input name.
     - level -> level of the required transfer.    
     - modelindex -> Index of the model used in the extra model list. For -1 value, the mean model is used.
     - parallel -> if True, the computation will be processed in parallel mode
    
    return depending on level value :
        - 0 -> return a list of the best approximation of the required result
        - 1 -> return a list of the absolute minimum value, best value, absolute maximum value
        - 2 -> return a list of the list of (minimum, value, maximum) for each extra model
    """
    maxworkers = -2
    max_workers = maxWorkers(maxworkers)
    if parallel and max_workers > 1:
        result = [None for _ in modellist]
        with cf.ProcessPoolExecutor(max_workers=max_workers) as executor: 
            futures = [executor.submit(_usageJob, ind, 
                sublist, inputs, level, modelindex) 
                for ind, sublist in enumerate(modellist)]
            for indloc, future in enumerate(cf.as_completed(futures)):
                try:
                    ind, modname, res = future.result()
                except:
                    res = (0,0,0)
                    modname = ""
                    ind =indloc
                result[ind] = modname, res 
                if progress:
                    progress(indloc)           
    else:
        result = []
        for ind, sublist in enumerate(modellist):
            indloc, modname, res = __usageJob(ind, sublist, 
                                             inputs, level, modelindex)
            result.append((modname, res))
            if progress:
                progress(indloc)           
    return result

def _usageJob(ind, sublist, inputs, level, modelindex):
    (sourcefile, modelname, smiles) = sublist[:3]
    if smiles is None:
        res = (0,0,0) if level else 0
    else:
        try:
            model = get_model(sourcefile, modelname=modelname, smiles=smiles)
            res = get_result(model, inputs, level, modelindex)
        except Exception as err:
            res = (0,0,0) if level else 0
    return ind, modelname, res

def __usageJob(ind, sublist, inputs, level, modelindex):
    (sourcefile, modelname, smiles) = sublist[:3]
    if smiles is None:
        res = (0,0,0) if level else 0
    else:
        try:
            model = get_model(sourcefile, modelname=modelname, smiles=smiles)
            res = get_result(model, inputs, level, modelindex)
        except:
            res = (0,0,0) if level else 0
    return ind, modelname, res

def get_model(sourcefile, modelname="", smiles="", name_smiles_list=[]):
    """Creation of a model from a model file. This model may be as well
    a neural network model as a graphmachine model, depending upon the
    model file provided. In case of a graphmachine model, the parameter
    'smiles' is mandatory.
    
    parameters:
     - sourcefile -> file created at the end of the training process (extensions may be 'nnx', 'gmx' or 'nml' for upstream compatibility).
     - modelname -> name of the molecule you want to apply the model to optional parameter.
     - smiles -> smiles code of the molecule you want to apply the model to. Mandatory parameter.
     - name_smiles_list -> iterable of tuples (name, smiles). If name_smiles_list is not empty, it overwrite "modelname" and "smiles" arguments.
    
    If name_smiles_list is not empty, return a list of graphmachine model.
    If name_smiles_list is empty, return the created graphmachine model,
    or None if impossible.
    
    The model has many interesting properties:
     - propertyName -> Name of the current property
     - modelName -> Name of the model
     - baselen -> Length of the training base
     - extraCount -> Number of side models included
     - inputCount -> Number of inputs
     - weights -> Current weights vector
     - paramCount -> Number of weights
     - dispersionMatrix -> Current dispersion matrix
     
    """
    if name_smiles_list is not None and len(name_smiles_list):
        return [get_model(sourcefile, modelname, smiles) for modelname, smiles in name_smiles_list]
    model = None
    modeldata = None
    innames = []
    outames = []
    if smiles:
        model = createGMModel(sourcefile, modelname, smiles)
        return model
    dico = getDictFromGmx(sourcefile)
    if "xml" in dico.keys():
        xml = dico["xml"]
        model = Driver(xml)
    elif "modeldata" in dico.keys():
        modeldata = dico['modeldata']
    elif "model" in dico.keys():
        modeldict = dico['model']
        innames = modeldict["inputnames"].split(',')
        ni = len(innames)
        outames = modeldict["outputname"].split(',')
        no = len(outames)
        nh = int(modeldict["hidden"])
        id = modelname
        modeldata = ModelData(input=ni, output=no, hidden=nh, name=id)
    if modeldata:
        model = Driver(modeldata=modeldata)
        model.loadFromDict(dico)
    return model

def get_inputs(model):
    """Read the current input values of a model.
    
    parameters:
     - model -> model to read, obtained with 'get_model' function.
    """
    return [value for value in model.inputs]

def get_result(model, inputs=None, level=0, modelindex=-1):
    """Obtain the result of model transfer function.
    
    Parameters:
     - model -> model obtained with 'get_model' function. May be a list of models.
     - inputs -> vector, dictionary or None
         - vector : vector of the input values, in the order of the inputs.
         - dictionary: input value for each input name. This dictionary may include a level item. If it exists, this value will overwrite the explicit level value.
         - None : no new input values (default).
     - level -> level of the required transfer.    
     - modelindex -> Index of the model used in the extra model list. For -1 value, the mean model is used.
    
    return depending on level value :
        - 0 -> return the best approximation of the required result
        - 1 -> return absolute minimum value, best value, absolute maximum value
        - 2 -> return list of (minimum, value, maximum) for each extra model
    """
    if isinstance(model, (list, tuple)):
        ## mettre en parallel ?!?
        return [get_result(mod, inputs, level, modelindex) for mod in model]
    if isinstance(inputs, dict):
        if 'level' in inputs:
            # On ne supprime pas la cle 'level' du dictionnaire.
            level = inputs['level']
        if hasattr(model, 'inputNames'):
            namegen = model.inputNames
        else:
            namegen = model.mainModel.inputNames
        inpts = [inputs[inname] for inname in namegen]
        inputs = inpts    
    res = model.transferEx(inputs, withCI=2, modelindex=modelindex)
    outs = [float(val) for val in res[0]]
    CIps = [float(val) for val in res[1]]
    CIms = [float(val) for val in res[2]]
    levs = [float(val) for val in res[3]]
    if modelindex == -1:
        outmean = reduce(lambda x, y: x + y, outs) / float(len(outs))
        minCIms = min(CIms)
        maxCIps = max(CIps)
    else:
#         outmean = outs[modelindex]
#         minCIms = CIms[modelindex]
#         maxCIps = CIps[modelindex]
        outmean = np.squeeze(outs)
        minCIms = np.squeeze(CIms)
        maxCIps = np.squeeze(CIps)
    model.lastoutput = outmean
    if not level:
        return outmean
    if level == 1:
        return minCIms, outmean, maxCIps

    return outs, CIms, CIps, levs

def get_reverse_result(model, target, inputs=None, fixedinputs=None,
        level=0 , epochs=None, modelindex=-1, callback=None, debug=0): 
    """**Only for standard neural network models.**
    
    Run a reverse training, modifying the inputs in order to fit the model output to the target. 
    
    Parameters:    
     - model -> model obtained with 'get_model' function.
     - target -> output value target.
     - inputs -> initial input values. It may be a vector-like, a dictionary or None:
         - vector-like : vector of the initial input values, in the order of the inputs.
         - dictionary: initial input value for each input name.
         - None : no imposed initial input values (default).
     - fixedinputs -> fixed input values.  It may be a vector-like, a dictionary or None:
         - vector-like : vector of boolean values, in the order of the inputs.
         - dictionary : each input may be fixed (or not) by its name. A missing input is considered as free.
         - None : all input values are free
     - epochs -> maximum number of optimizing computation epochs. Defaulted to 100.
     - callback -> callback function. It takes a current index as parameter. Default None
     - modelindex -> Index of the model used in the extra model list. For -1 value, the mean model is used.
     
    return tuple (newinputs, code):
     - newinputs -> the new set of input values
     - code -> reverse_train return code :
         - code & 1 : an unacceptable point has been ontained
         - code & 2 : accuracy limot has been reached
         - code & 4 : epochs limit has been reached
    """
    if epochs is None or epochs <= 0:
        epochs = 100
    newinputs, _ = model.reverse_train(target, inputs, fixedinputs,
                                       epochs, modelindex, callback, 
                                       debug)
    res = get_result(model, newinputs, level, modelindex)
    return res

if __name__ == '__main__':
    source = '/Users/jeanluc/Documents/Workspace/Monal_Test/src/testFiles/GM_B13C81Acetophenones_50T_0I_150E_10S_SEED1947_5N.nnx'
    modellist = [(source, "2,2-dichloroacetophenone0" , "c(cc1)cc[c:1]1C(=O)C(Cl)Cl", None),
(source, "2,2-dichloroacetophenone1", "c(c[c:1]1)ccc1C(=O)C(Cl)Cl", None),
(source, "2,2-dichloroacetophenone2", "c([c:1]c1)ccc1C(=O)C(Cl)Cl", None),
(source, "2,2-dichloroacetophenone3", "[c:1](cc1)ccc1C(=O)C(Cl)Cl", None),
(source, "4'-butoxyacetophenone0", "CCCCO[c:1]1ccc(cc1)C(=O)C", None),
(source, "4'-butoxyacetophenone1", "CCCCOc1ccc(c[c:1]1)C(=O)C", None),
(source, "4'-butoxyacetophenone2", "CCCCOc1ccc([c:1]c1)C(=O)C", None),
(source, "4'-butoxyacetophenone3", "CCCCOc1cc[c:1](cc1)C(=O)C", None),
(source, "2-bromoacetophenone0", "BrCC(=O)[c:1]1ccccc1", None),
(source, "2-bromoacetophenone1", "BrCC(=O)c1cccc[c:1]1", None),
(source, "2-bromoacetophenone2", "BrCC(=O)c1ccc[c:1]c1", None),
(source, "2-bromoacetophenone3", "BrCC(=O)c1cc[c:1]cc1", None),]
    
    res = get_results_from_smiles(modellist, None, level=1, modelindex=-1, parallel=True)
    for mname, val in res:
        st = "{0} : \t{1}".format(mname, str(val))
        print(st)
#        print(mname, "->", val)
    
    
    
    
    
#     inputs = [28, 11, 18, 79, 17]
#     model = get_model(source)
#     model.reprdepth = 1
#     print(repr(model))
#     dico = {
#         'X1': 28.0,
#         'X2': 11.0,
#         'X3': 18.0,
#         'X4': 79.0,
#         'X5': 17.0,
#     }
#     modelindex = 1
#     res = get_result(model, dico, modelindex=modelindex)
#     print(model.lastoutput)
#     print(res)
#     print()
#     target = 25
#     print("target", target)
#     fixed = None
# #     fixed = [1,1,0,1,1]
#     epochs = None
#     res = get_reverse_result(model, target, fixedinputs=fixed, epochs=epochs, modelindex=modelindex, debug=1)
#     newinputs = get_inputs(model)
#     print(newinputs)
#     print("result", res)
#     print("residu", res - target)

# 2,2,3-trimethylbutane    CC(C)(C)C(C)C

#     sourcegm = "/Users/jeanluc/Documents/Workspace/Monal_Test/src/testFiles/GM_B13C81Acetophenones_50T_0I_150E_10S_SEED1947_5N.nnx"
#     modelname = "acetophenone0"
#     smile = "c(cc1)cc[c:1]1C(=O)C"
#     pair0 = ("acetophenone0", "[c:1](cc1)ccc1C(=O)C")
#     pair1 = ("acetophenone1", "c(cc1)cc[c:1]1C(=O)C")
#     pair2 = ("2,2,3-trimet
#     modelgms = get_model(sourcegm, name_smiles_list=(pair0, pair1))
#     valuegms = get_result(modelgms, level=1)
#     for val in valuegms:
#         print("gm value", val)
#     print("gm value", valuegm)
#     model = get_model(source, 'unknown')
#     print(model.name, model.propertyName)
#
# # call with level 0
#     print("mean best value:")
#     value = get_result(model, inputs)
#
#     print("\t{0:9.6f}".format(value))
#     print()
#     titles = ["min", "value", "max"]
#     print("\t".join(titles))
#
# # call with level 1
#     mini, val, maxi = get_result(model, inputs, level=1)
#
#     values = ["{0:6.3f}".format(val) for val in [mini, val, maxi]]
#     print("\t".join(values))
#     print()
#     print("available results")
#     titles = ["value", "max", "min", "leverage"]
#     print("\t".join(titles))
#
# # call with level 2
#     outs, CIps, CIms, levs = get_result(model, inputs, level=2)
#     for out, CIp, CIm, lev in zip(outs, CIps, CIms, levs):
#         res = ["{0:6.3f}".format(val) for val in [out, CIp, CIm, lev]]
#         st = "\t".join(res)
#         print(st)
    print('Done')